/*******************************************************************************
 * OSUG-DOI project ( http://doi.osug.fr ) - Copyright (C) CNRS.
 ******************************************************************************/
package fr.osug;

import ch.qos.logback.classic.Level;
import fr.osug.doi.ProcessPipeline;
import fr.osug.doi.DOIConfig;
import fr.osug.doi.Paths;
import fr.osug.doi.ProjectConfig;
import java.io.IOException;
import java.nio.charset.Charset;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import java.util.TimeZone;
import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.ApplicationArguments;
import org.springframework.boot.ApplicationRunner;
import org.springframework.boot.ExitCodeGenerator;
import org.springframework.boot.SpringApplication;
import org.springframework.boot.autoconfigure.SpringBootApplication;
import org.springframework.context.annotation.Profile;
import org.springframework.stereotype.Component;
import org.springframework.transaction.annotation.Transactional;

@SpringBootApplication
public class DOIApplication {

    private final static Logger logger = LoggerFactory.getLogger(DOIApplication.class.getName());

    /**
     * Main entry point
     * @param args command-line arguments
     * @throws Exception 
     */
    public static void main(final String[] args) throws Exception {
        // return exitCode from ExitCodeGenerator:
        System.exit(
                SpringApplication.exit(
                        SpringApplication.run(DOIApplication.class, args)
                )
        );
    }

    // members:
    @Autowired
    private DOIConfig doiConfig;

    @PostConstruct
    public void initialize() {
        // Set the default locale to en-US locale (for Numerical Fields "." ",")
        Locale.setDefault(Locale.US);

        TimeZone.setDefault(TimeZone.getTimeZone("UTC"));

        logger.info("Default Charset=" + Charset.defaultCharset());

        // mail throw IllegalStateException:
        final String basePath = Paths.DIR_BASE;
        logger.info("Base Path=" + basePath);
    }

    @PreDestroy
    public void onExit() {
        if (doiConfig.isDebug() && logger.isInfoEnabled()) {
            logger.info("Exit hook: components provided by Spring Boot:");

            final String[] beanNames = doiConfig.getAppCtx().getBeanDefinitionNames();
            Arrays.sort(beanNames);

            for (String beanName : beanNames) {
                logger.info(beanName);
            }
        }
    }

    @Component
    @Profile("!test")
    public static class DOIPipelineRunner implements ApplicationRunner, ExitCodeGenerator {

        private final static String ARG_ACTION = "action";

        // members:
        @Autowired
        private DOIConfig doiConfig;
        /* exit status */
        private int exitCode = 0;

        @Override
        public void run(final ApplicationArguments aa) {
            // Just leave method if no argument has been given
            if (aa.getSourceArgs().length == 0) {
                fail("Missing arguments !");
            }

            if (aa.containsOption("v")) {
                final String levelString = getSingleValue("v", aa);
                setLoggerLevel(LoggerFactory.getLogger("fr.osug.doi"), levelString);
            }

            if (logger.isInfoEnabled()) {
                for (String opt : aa.getOptionNames()) {
                    logger.info("argument [{}] = '{}'", opt, aa.getOptionValues(opt));
                }
            }

            final String action = getRequiredValue(ARG_ACTION, aa);
            switch (action) {
                case "help":
                    showArgumentsHelp();
                    break;
                case "process":
                    doProcess(aa);
                    break;
                default:
                    if (action != null) {
                        fail("Unsupported action: " + action);
                    }
            }
        }

//        @Transactional
        public void doProcess(ApplicationArguments aa) {

            // Required values:
            final String project = getRequiredValue("project", aa);

            final boolean saveCSV = aa.containsOption("csv");

            try {
                final ProjectConfig projectConfig = new ProjectConfig(project);
                projectConfig.setSaveCSV(saveCSV);

                final ProcessPipeline pipeline = new ProcessPipeline(
                        doiConfig,
                        projectConfig);

                pipeline.process();

            } catch (IOException ioe) {
                // will rollback transaction:
                throw new RuntimeException(ioe);
            }
        }

        /** Show command arguments help. */
        public static void showArgumentsHelp() {
            System.out.println("---------------------------------- Arguments help ------------------------------");
            System.out.println("| Key          Value           Description                                     |");
            System.out.println("|------------------------------------------------------------------------------|");
            System.out.println("| --action=[help|process]       Define the command to perform                  |");
            System.out.println("| [--v=0|1|2|3|4|5]             Define console logging level                   |");
            System.out.println("|                                                                              |");
            System.out.println("| LOG LEVELS : 0 = OFF, 1 = ERROR, 2 = WARNING, 3 = INFO, 4 = DEBUG, 5 = ALL   |");
            System.out.println("|------------------------------------------------------------------------------|");
            System.out.println("| Action [process]:                                                            |");
            System.out.println("| --project=<dir>               Project folder name (case sensitive)           |");
            System.out.println("| [--csv]                       CSV Output format                              |");
            System.out.println("| [--xml]                       XML Output format (default)                    |");
            System.out.println("|------------------------------------------------------------------------------|\n");
        }

        private void fail(String msg) {
            showArgumentsHelp();
            setExitCode(-1);
            throw new IllegalArgumentException(msg);
        }

        @Override
        public int getExitCode() {
            return exitCode;
        }

        private void setExitCode(int exitCode) {
            this.exitCode = exitCode;
        }

        private String getRequiredValue(String opt, final ApplicationArguments aa) {
            if (!aa.containsOption(opt)) {
                fail("Missing argument '" + opt + "' !");
                return null;
            }
            return getSingleValue(opt, aa);
        }

        private static String getSingleValue(String opt, final ApplicationArguments aa) {
            final List<String> values = aa.getOptionValues(opt);
            if (values.size() != 1) {
                throw new IllegalArgumentException("Too many values for argument '" + opt + "': " + values);
            }
            return values.get(0);
        }

        /**
         * Set given logger's level (if it is a logback logger)
         * @param logger slf4j logger
         * @param levelString level as string
         */
        private static void setLoggerLevel(Logger logger, String levelString) {
            Level level = null;
            switch (levelString) {
                case "0":
                    level = Level.OFF;
                    break;
                case "1":
                    level = Level.ERROR;
                    break;
                case "2":
                    level = Level.WARN;
                    break;
                case "3":
                    level = Level.INFO;
                    break;
                case "4":
                    level = Level.DEBUG;
                    break;
                case "5":
                    level = Level.ALL;
                    break;
                default:
                    logger.info("Invalid Level [{}]", level);
                    showArgumentsHelp();
                    return;
            }
            if (logger instanceof ch.qos.logback.classic.Logger) {
                logger.info("Set logger[{}] Level to [{}]", logger.getName(), level);
                ((ch.qos.logback.classic.Logger) logger).setLevel(level);
            }
        }

    }

}
