/*******************************************************************************
 * OSUG-DOI project ( http://doi.osug.fr ) - Copyright (C) CNRS.
 ******************************************************************************/
package fr.osug.doi;

import fr.osug.doi.domain.Doi;
import fr.osug.doi.domain.DoiStaging;
import fr.osug.doi.domain.Project;
import fr.osug.doi.domain.IndexEntry;
import fr.osug.doi.domain.NameEntry;
import fr.osug.doi.repository.DoiStagingRepository;
import fr.osug.doi.repository.ProjectRepository;
import fr.osug.doi.service.DoiService;
import fr.osug.doi.validation.ValidationUtil;
import fr.osug.util.FileUtils;
import fr.osug.xml.validator.ErrorMessage;
import fr.osug.xml.validator.ValidationResult;
import fr.osug.xml.validator.XmlValidator;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.transaction.annotation.Transactional;

/**
 *
 * @author bourgesl
 */
public final class ProcessPipeline {

    private final static Logger logger = LoggerFactory.getLogger(ProcessPipeline.class.getName());

    public final static String HTML_EXT = ".html";
    public final static String HTML_INDEX = "index" + HTML_EXT;
    public final static String HTML_REPORT = "report" + HTML_EXT;
    public final static String HTML_ERROR = "error" + HTML_EXT;

    public final static String XSD_DOI_SCHEMA = "file://" + Paths.DIR_XSD_SCHEMA + "/metadata.xsd";

    private final static String XSLT_XML_TO_DOI = "xml2doi.xsl";
    private final static String XSLT_DOI_TO_LANDING_PAGE = "doi2landing.xsl";

    private final static String XSLT_PROJECT_INDEX = "project_index.xsl";
    private final static String XSLT_PROJECT_REPORT = "project_report.xsl";

    private final static String CONFIG_ACCESS_INSTRUCTIONS = "access_instruction.html";

    // xslt parameters (common):
    private final static String XSLT_PARAM_DATE = "date";
    private final static String XSLT_PARAM_WEB_ROOT = "www_root";

    // xslt parameters (doi2landing):
    private final static String XSLT_PARAM_DATA_ACCESS_HTML = "data_access_html";
    private final static String XSLT_PARAM_DATA_ACCESS_URL = "data_access_url";
    private final static String XSLT_PARAM_EMBEDDED = "embedded";
    private final static String XSLT_PARAM_EMBEDDED_FULL = "full";
    private final static String XSLT_PARAM_EMBEDDED_HEADER = "header";
    private final static String XSLT_PARAM_EMBEDDED_META = "meta";
    private final static String XSLT_PARAM_LANDING_PAGE_URL = "landing_page_url";
    private final static String XSLT_PARAM_METADATA_FILE = "metadataFile";

    // xslt parameters (index_project):
    private final static String XSLT_PARAM_PARENT = "parent";
    private final static String XSLT_PARAM_CURRENT = "current";

    // members
    private Date _now = null;
    private final DOIConfig doiConfig;
    private final ProjectConfig projectConfig;
    // temporary data
    private final PipelineGlobalData globalData = new PipelineGlobalData();

    public ProcessPipeline(final DOIConfig doiConfig,
                           final ProjectConfig projectConfig) throws IOException {

        this.doiConfig = doiConfig;
        this.projectConfig = projectConfig;
    }

    public void process() throws IOException {

        // 1 - Process complete datasets (csv):
        for (File inputCsvFile : projectConfig.getMetadataCsvFiles()) {
            processCsvFile(inputCsvFile, false);
        }

        // 2 - Process partial datasets (csv to be merged with templates):
        for (File inputCsvFile : projectConfig.getInputCsvFiles()) {
            processCsvFile(inputCsvFile, true);
        }

        // TODO: 3 - process complete datasets as XML files (full metadata)
        // global validation of references
        validateRefs();

        // 4 - Landing pages
        generateLandingPages();

        // TODO: test publication datacite ?
        // Update database:
        updateDatabase();

        // Update index pages:
        generateIndexPagesStaging();
    }

    private void generateLandingPages() throws IOException {

        final File webStaging = projectConfig.getWebStagingProjectDir();
        logger.info("Generating landing pages into {}", webStaging);

        final File webStagingEmbed = projectConfig.getWebStagingProjectEmbedDir();
        final File webStagingXml = projectConfig.getWebStagingProjectXmlDir();

        final File dataAccessFileLocation = new File(projectConfig.getProjectConf(), CONFIG_ACCESS_INSTRUCTIONS);
        final File dataAccessFile = FileUtils.getFile(dataAccessFileLocation);
        if (dataAccessFile == null) {
            logger.warn("Missing dataAccessFile: {}", dataAccessFileLocation);
        }

        // TODO: use mapping file [DOI:dataAccessUrl] like url mapping:
        final String defaultDataAccessUrl = projectConfig.getProperty(ProjectConfig.CONF_KEY_DATA_ACCESS_URL);

        final Map<String, Object> xslParameters = new HashMap<String, Object>(8);
        xslParameters.put(XSLT_PARAM_DATE, now().toString());
        if (dataAccessFile != null) {
            xslParameters.put(XSLT_PARAM_DATA_ACCESS_HTML, dataAccessFile.toURI().toString());
        }
        if (defaultDataAccessUrl != null) {
            xslParameters.put(XSLT_PARAM_DATA_ACCESS_URL, defaultDataAccessUrl);
        }

        for (PipelineDoiData doiData : globalData.getDoiDatas().values()) {
            final String doiSuffix = doiData.getDoiSuffix();
            logger.info("processing {}", doiSuffix);

            // Get external Landing page URL:
            final String landingPageUrl = projectConfig.getLandingPageUrl(doiSuffix);
            if (landingPageUrl != null) {
                logger.info("landingPageUrl: {}", landingPageUrl);
                xslParameters.put(XSLT_PARAM_LANDING_PAGE_URL, landingPageUrl);

                doiData.setLandingPageUrl(landingPageUrl);
            } else {
                xslParameters.remove(XSLT_PARAM_LANDING_PAGE_URL);
            }

            // add warnings:
            if (dataAccessFile == null) {
                doiData.addError("Missing data access instructions (" + CONFIG_ACCESS_INSTRUCTIONS + ")");
            }
            if (StringUtils.isEmpty(defaultDataAccessUrl)) {
                doiData.addError("Missing data access url");
            }

            final File stagingFileDOI = doiData.getStagingFileDOI();

            if (stagingFileDOI == null) {
                logger.info("Ignoring {} (no xml)", doiSuffix);
            } else {
                final String filenameNoExt = FileUtils.getFileNameWithoutExtension(stagingFileDOI);

                // main landing page:
                xslParameters.remove(XSLT_PARAM_EMBEDDED);
                xslParameters.put(XSLT_PARAM_WEB_ROOT, "../../");
                File output = new File(webStaging, filenameNoExt + HTML_EXT);
                generateLandingPage(xslParameters, stagingFileDOI, output);

                if (projectConfig.getPropertyBoolean(ProjectConfig.CONF_KEY_GENERATE_EMBEDDED)) {
                    // embedded mode:
                    xslParameters.put(XSLT_PARAM_WEB_ROOT, "../../../");

                    // embedded mode - full:
                    output = new File(webStagingEmbed, filenameNoExt + HTML_EXT);
                    xslParameters.put(XSLT_PARAM_EMBEDDED, XSLT_PARAM_EMBEDDED_FULL);
                    generateLandingPage(xslParameters, stagingFileDOI, output);

                    // embedded mode - header:
                    output = new File(webStagingEmbed, filenameNoExt + "-header" + HTML_EXT);
                    xslParameters.put(XSLT_PARAM_EMBEDDED, XSLT_PARAM_EMBEDDED_HEADER);
                    generateLandingPage(xslParameters, stagingFileDOI, output);

                    // embedded mode - meta:
                    output = new File(webStagingEmbed, filenameNoExt + "-meta" + HTML_EXT);
                    xslParameters.put(XSLT_PARAM_EMBEDDED, XSLT_PARAM_EMBEDDED_META);
                    generateLandingPage(xslParameters, stagingFileDOI, output);
                }

                // Xml DOI:
                output = new File(webStagingXml, stagingFileDOI.getName());
                FileUtils.copy(stagingFileDOI, output);
            }
        }
    }

    private void generateLandingPage(final Map<String, Object> xslParameters,
                                     final File stagingFileDOI, final File output) {

        logger.info("generateLandingPage: {}", output);

        xslParameters.put(XSLT_PARAM_METADATA_FILE, stagingFileDOI.getName());
        logger.debug("xslParameters: {}", xslParameters);

        // use XSLT to generate landing page from the doi xml file:
        doiConfig.getXmlFactory().transform(
                new StreamSource(stagingFileDOI),
                Paths.DIR_XSL + XSLT_DOI_TO_LANDING_PAGE,
                xslParameters, true,
                new StreamResult(output)
        );
    }

    @Transactional
    /* public needed by Transactional */
    public void updateDatabase() {
        final String projectName = projectConfig.getProjectName();

        final Date now = now();

        // Get or create the project:
        final DoiService doiService = doiConfig.getDoiService();
        final Project project = doiService.getOrCreateProject(projectName, now);

        // Update Dois:
        for (PipelineDoiData doiData : globalData.getDoiDatas().values()) {
            // Update DOI in staging phase:
            doiService.createOrUpdateStagingDoi(project, doiData.getDoiSuffix(),
                    doiData.getTitle(), doiData.getLandingPageUrl(),
                    doiData.isValid(), doiData.getMetadataMd5(),
                    doiData.messagesToString(), now);
        }

        // Update Project:
        doiService.updateProjectDate(projectName, now);
    }

    @Transactional
    /* public needed by Transactional */
    public void generateIndexPagesStaging() {
        final File webStaging = projectConfig.getWebStagingDir();

        final DoiService doiService = doiConfig.getDoiService();
        final ProjectRepository pr = doiService.getProjectRepository();
        final DoiStagingRepository dsr = doiService.getDoiStagingRepository();

        final List<Project> projects = pr.findAllByOrderByNameAsc();

        // Project listing:
        List<IndexEntry> entries = new ArrayList<IndexEntry>();

        for (Project p : projects) {
            final String projectName = p.getName();
            final Long countDoi = dsr.countByProject(projectName);
            final IndexEntry ie = new IndexEntry(projectName, p.getDescription(), p.getUpdateDate(), countDoi);

            logger.debug("IndexEntry: {}", ie);
            entries.add(ie);
        }
        // serialize entries and transform in the /staging/index.html
        IndexUtil.write(entries, globalData.buffer);

        String xmlDoc = globalData.buffer.toString();
        logger.debug("xmlDoc(project index): \n{}", xmlDoc);

        final Map<String, Object> xslParameters = new HashMap<String, Object>(8);
        xslParameters.put(XSLT_PARAM_DATE, now().toString());
        xslParameters.put(XSLT_PARAM_CURRENT, "staging");
        xslParameters.put(XSLT_PARAM_WEB_ROOT, "../");
        File output = new File(webStaging, HTML_INDEX);

        // use XSLT to generate the index page:
        doiConfig.getXmlFactory().transform(
                new StreamSource(new StringReader(xmlDoc)),
                Paths.DIR_XSL_ROOT + XSLT_PROJECT_INDEX,
                xslParameters, true,
                new StreamResult(output)
        );

        // DOI listing per project:
        for (Project p : projects) {
            final String projectName = p.getName();
            final List<DoiStaging> dsList = dsr.findByProject(projectName);

            logger.debug("DoiStaging list for project[{}]: {}", projectName, dsList);

            // index.html:
            entries.clear();

            for (DoiStaging ds : dsList) {
                logger.debug("DoiStaging: {}", ds);

                final Doi doi = ds.getDoi();

                final IndexEntry ie = new IndexEntry(doi.getIdentifier(), doi.getDescription(), ds.getUpdateDate());

                logger.debug("IndexEntry: {}", ie);
                entries.add(ie);
            }
            // serialize entries and transform in the /staging/PROJECT/index.html
            IndexUtil.write(entries, globalData.buffer);

            xmlDoc = globalData.buffer.toString();
            logger.debug("xmlDoc(doi index): \n{}", xmlDoc);

            xslParameters.put(XSLT_PARAM_PARENT, "staging");
            xslParameters.put(XSLT_PARAM_CURRENT, projectName);
            xslParameters.put(XSLT_PARAM_WEB_ROOT, "../../");
            // suppose child directory is the project folder:
            output = new File(new File(webStaging, projectName), HTML_INDEX);

            // use XSLT to generate the index page:
            doiConfig.getXmlFactory().transform(
                    new StreamSource(new StringReader(xmlDoc)),
                    Paths.DIR_XSL_ROOT + XSLT_PROJECT_INDEX,
                    xslParameters, true,
                    new StreamResult(output)
            );

            // serialize doi staging list and transform in the /staging/PROJECT/report.html
            IndexUtil.writeReport(dsList, globalData.buffer);

            xmlDoc = globalData.buffer.toString();
            logger.debug("xmlDoc(doi report): \n{}", xmlDoc);

            xslParameters.put(XSLT_PARAM_PARENT, "staging");
            xslParameters.put(XSLT_PARAM_CURRENT, projectName);
            xslParameters.put(XSLT_PARAM_WEB_ROOT, "../../");
            // suppose child directory is the project folder:
            output = new File(new File(webStaging, projectName), HTML_REPORT);

            // use XSLT to generate the report page:
            doiConfig.getXmlFactory().transform(
                    new StreamSource(new StringReader(xmlDoc)),
                    Paths.DIR_XSL_ROOT + XSLT_PROJECT_REPORT,
                    xslParameters, true,
                    new StreamResult(output)
            );

            // serialize doi staging error list and transform in the /staging/PROJECT/error.html
            final List<DoiStaging> dsErrorList = dsr.findErrorByProject(projectName);
            logger.debug("DoiStaging error list for project[{}]: {}", projectName, dsErrorList);

            IndexUtil.writeReport(dsErrorList, globalData.buffer);

            xmlDoc = globalData.buffer.toString();
            logger.debug("xmlDoc(error report): \n{}", xmlDoc);

            xslParameters.put(XSLT_PARAM_PARENT, "staging");
            xslParameters.put(XSLT_PARAM_CURRENT, projectName);
            xslParameters.put(XSLT_PARAM_WEB_ROOT, "../../");
            // suppose child directory is the project folder:
            output = new File(new File(webStaging, projectName), HTML_ERROR);

            // use XSLT to generate the report page:
            doiConfig.getXmlFactory().transform(
                    new StreamSource(new StringReader(xmlDoc)),
                    Paths.DIR_XSL_ROOT + XSLT_PROJECT_REPORT,
                    xslParameters, true,
                    new StreamResult(output)
            );
        }
        logger.info("generateIndexPagesStaging: done.");
    }

    private void validateRefs() throws IOException {
        logger.debug("validateRefs");

        final DoiService doiService = doiConfig.getDoiService();
        final String publicPrefix = doiConfig.getPrefix();

        for (String doiId : globalData.getDoiIds()) {
            logger.debug("checking refs for [{}]", doiId);

            final Set<String> refs = globalData.getDoiRefs(doiId);
            if (refs != null) {
                for (String refId : refs) {
                    // Only check DOI references with Prefix Test or Public:
                    if (ValidationUtil.isTestPrefix(refId) || refId.startsWith(publicPrefix)) {
                        logger.debug("validateRefs: check ref[{}]", refId);

                        boolean found = false;

                        // Lookup first in processed identifiers:
                        if (globalData.hasDoiId(refId)) {
                            found = true;
                        } else {
                            // Then in database (previous data):
                            final String doiSuffix = DoiCsvData.getDoiSuffix(refId);
                            if (doiService.getDoi(doiSuffix) != null) {
                                found = true;
                            }
                        }
                        if (!found) {
                            final PipelineDoiData doiData = globalData.getDoiData(doiId);
                            if (doiData != null) {
                                doiData.addError("Invalid " + Const.KEY_REL_ID_START + " '" + refId + "' found.");
                            }
                        }
                    }
                }
            }
        }

        // Check name references:
        final TreeSet<NameEntry> entries = globalData.getNameRefs();
        if (logger.isDebugEnabled()) {
            logger.debug("name references:\n{}", entries);
        }

        // Collect unique attributes:
        final Set<String> keys = new LinkedHashSet<String>();
        for (NameEntry e : entries) {
            if (e.hasAttributes()) {
                for (String k : e.getAttributes().keySet()) {
                    keys.add(k);
                }
            }
        }

        final int nCols = keys.size() + 1;

        final List<String[]> rows = new ArrayList<String[]>();
        // header:
        String[] cols = new String[nCols];
        int i = 0;
        cols[i++] = "#Name";
        for (String k : keys) {
            cols[i++] = k;
        }
        rows.add(cols);
        // data:
        for (NameEntry e : entries) {
            if (e.hasAttributes()) {
                cols = new String[nCols];
                i = 0;
                cols[i++] = e.getName();
                for (String k : keys) {
                    cols[i++] = e.getAttribute(k);
                }
            } else {
                cols = new String[]{e.getName()};
            }
            rows.add(cols);
        }
        final CsvData data = new CsvData(rows);
        if (logger.isDebugEnabled()) {
            logger.debug("name reference table:\n{}", data);
        }

        // Save file as CSV:
        final File fileCSV = new File(projectConfig.getTmpDir(), "names" + Const.FILE_EXT_CSV);
        CsvUtil.write(data, fileCSV);
    }

    private void processCsvFile(final File inputCsvFile,
                                final boolean doMerge) throws IOException {
        /*
        Spliter et produire 1 fichier XML dataCite par jeu.
        Ex: AMMA-CATCH.CL.Run_O.xml
         */
        final List<DoiCsvData> csvDatas = DoiCsvSplitter.split(inputCsvFile);

        for (DoiCsvData data : csvDatas) {
            if (logger.isDebugEnabled()) {
                logger.debug("process: data:\n{}", data);
            }
            if (!data.isEmpty()) {
                final String doiId = data.getDoiId();

                // check identifier (missing) ?
                if (ValidationUtil.checkIdentifier(doiId, data)) {
                    final String doiSuffix = data.getDoiSuffix();
                    logger.info("process: dataset[{}].", doiSuffix);

                    final PipelineDoiData doiData = globalData.newDoiData(doiId, doiSuffix, data.getTitle());

                    // ignore duplicates
                    if (doiData != null) {
                        // validate identifier (format)
                        ValidationUtil.validateIdentifier(doiId, doiData);

                        if (doMerge) {
                            mergeCSV(data, doiData);
                        }

                        // Validate metadata:
                        validateData(data, doiData);

                        // Save metadata for dataset:
                        saveData(data, doiData);
                    }
                }
            }
        }
    }

    private void mergeCSV(final DoiCsvData data, final PipelineDoiData doiData) {

        final DoiTemplates templates = projectConfig.getTemplates();

        /*                
        ### 2/ Fusionner avec template observatoire : template_obs.csv

        Informations disjointes
         */
        DoiCsvData t = templates.getBase();
        if (t == null) {
            doiData.addError("Missing base template for project '" + projectConfig.getProjectName() + "'");
        } else {
            data.mergeFrom(t);
        }

        final String doiId = data.getDoiId(); // not null

        /*
        ### 3/ Traitement spécifiques: fusionner avec template pays

        Fusionner avec template pays (template_benin.csv):

        Trouver le template du pays qui contient geoLocationPlace;Benin
         */
        final String geoLocationPlace = data.getFirstGeoLocationPlace();
        if (geoLocationPlace == null) {
            doiData.addWarning("Missing ''" + Const.KEY_GEO_LOCATION_PLACE + "'' in dataset '" + doiId + "'");
        } else {
            t = templates.getByGeoLocationPlace(geoLocationPlace);
            if (t == null) {
                doiData.addError("Missing country template for '" + geoLocationPlace + "'");
            } else {
                logger.info("Using country template '{}'", geoLocationPlace);
                data.mergeFrom(t);
            }
        }

        /*
        ### 4/ Traitement spécifiques: fusionner avec template jeu

        Trouver le nom du jeu dans le XML:
        <identifier identifierType="DOI">10.5072/AMMA-CATCH.CL.Run_O</identifier>

        Trouver le template du jeu qui contient identifier:DOI;10.5072/AMMA-CATCH.CL.Run_O
         */
        t = templates.getByIdentifier(doiId);
        if (t == null) {
            doiData.addWarning("Missing dataset template for id '" + doiId + "'");
        } else {
            logger.info("Using dataset template for id '{}'", doiId);
            data.mergeFrom(t);
        }
    }

    private void validateData(final DoiCsvData data, final PipelineDoiData doiData) {
        logger.debug("validateData [{}]", doiData.getDoiId());
        // TODO: check title ?

        // check all values:
        final List<String[]> rows = data.getRows();
        for (ListIterator<String[]> it = rows.listIterator(); it.hasNext();) {
            final String[] cols = it.next();
            if (cols != null && cols.length >= 2) {
                if (cols[1].toLowerCase().contains("todo")) {
                    doiData.addWarning("TODO detected for key [" + cols[0] + ']');
                }

                // Collect name references:
                final String key = cols[0];
                if (key.startsWith(Const.KEY_CREATOR_NAME) || key.startsWith(Const.KEY_CONTRIBUTOR_NAME)) {
                    globalData.addNameRef(cols);
                }
            }
        }

        final Set<String> refs = data.getReferences();
        if (refs != null) {
            logger.debug("References: {}", refs);
        }
        // anyway: add the doi entry:
        globalData.addRefs(doiData.getDoiId(), refs);
    }

    private void saveData(final DoiCsvData data, final PipelineDoiData doiData) throws IOException {
        // Sort keys:
        data.sort();

        // Get temporary directory:
        final File tmpDoiDir = projectConfig.getTmpDoiDir();
        // Get Staging directory:
        final File stagingDir = projectConfig.getStagingDir();

        // File base name:
        final String baseName = data.getDoiSuffix();

        // Save file as XML:
        DoiCsvToXml.write(data, globalData.buffer);

        final String xmlDoc = globalData.buffer.toString();
        logger.debug("xmlDoc: \n{}", xmlDoc);

        boolean forceSaveCSV = false;
        boolean forceSaveXML = false;

        // check non empty file:
        if (xmlDoc.length() == 0) {
            forceSaveCSV = true;
            doiData.addError("No XML output");
        } else {
            globalData.doiDoc.reset();

            // use XSLT to write doi xml file:
            doiConfig.getXmlFactory().transform(
                    new StreamSource(new StringReader(xmlDoc)),
                    Paths.DIR_XSL + XSLT_XML_TO_DOI, true,
                    new StreamResult(globalData.doiDoc)); // no xslt-param

            final String doiDoc = globalData.doiDoc.toString();
            logger.debug("doiDoc: \n{}", doiDoc);

            if (doiDoc.length() == 0) {
                forceSaveXML = true;
                doiData.addError("No Metadata XML output");
            } else {
                final File fileDOI = new File(tmpDoiDir, baseName + Const.FILE_EXT_XML);
                DoiCsvToXml.writeXml(doiDoc, fileDOI);

                // compute MD5 on metadata:
                final String xmlMd5 = FileUtils.MD5(new FileInputStream(fileDOI));
                doiData.setMetadataMd5(xmlMd5);
                logger.debug("DOI metadata [{}] MD5: {}", fileDOI, xmlMd5);

                // Validate:
                final XmlValidator validator = doiConfig.getXmlValidatorFactory().getInstance(XSD_DOI_SCHEMA);

                final ValidationResult vr = new ValidationResult();
                validator.validate(new StreamSource(fileDOI), vr);

                logger.info("XSD validation [{}]: {}", fileDOI.getAbsolutePath(), vr.isValid());

                if (!vr.isValid()) {
                    doiData.addError("Metadata XML validation failed:");

                    for (ErrorMessage msg : vr.getMessages()) {
                        doiData.addError(msg.toString());
                    }
                }

                // Anyway copy the doi metadata into staging:
                final File stagingFileDOI = new File(stagingDir, fileDOI.getName());

                logger.debug("Copy [{}] to [{}]", fileDOI, stagingFileDOI);

                FileUtils.copy(fileDOI, stagingFileDOI);

                // to be processed at next step:
                doiData.setStagingFileDOI(stagingFileDOI);
            }
        }

        // Always save intermediate results in case of failure:
        if (forceSaveCSV || projectConfig.isSaveCSV()) {
            // Save file as CSV:
            final File fileCSV = new File(tmpDoiDir, baseName + Const.FILE_EXT_CSV);
            CsvUtil.write(data, fileCSV);
        }

        if (forceSaveXML || doiConfig.isDebug()) {
            final File fileXML = new File(tmpDoiDir, baseName + "-xml" + Const.FILE_EXT_XML);
            DoiCsvToXml.writeXml(xmlDoc, fileXML);
        }
    }

    private Date now() {
        if (_now == null) {
            _now = new Date();
        }
        return _now;
    }
}
