/*******************************************************************************
 * OSUG-DOI project ( http://doi.osug.fr ) - Copyright (C) CNRS.
 ******************************************************************************/
package fr.osug.doi;

import fr.osug.util.FileUtils;
import java.io.File;
import java.io.IOException;
import java.io.Reader;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Properties;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author bourgesl
 */
public final class ProjectConfig {

    public final static String CONF_PROJECT = "project.properties";

    public final static String CONF_DOI_URL_MAPPING = "doi_url_landing_page.csv";

    public final static String CONF_KEY_DATA_ACCESS_URL = "dataAccessUrl";
    public final static String CONF_KEY_GENERATE_EMBEDDED = "generate_embedded";

    private final static File[] EMPTY = new File[0];

    private final static Logger logger = LoggerFactory.getLogger(ProjectConfig.class.getName());

    private final String projectName;
    private final File projectConf;

    private final Properties properties;

    private final Map<String, String> landingPageUrlMapping;

    private final DoiTemplates templates;
    private File[] metadataCsvFiles = EMPTY;
    private File[] inputCsvFiles = EMPTY;
    private File tmpDir;
    private File tmpDoiDir;
    private File stagingDir;
    private File webStagingDir;
    private File webStagingProjectDir;
    private File webStagingProjectEmbedDir;
    private File webStagingProjectXmlDir;
    private File webPublicDir;
    private File webPublicProjectDir;

    private boolean saveCSV = false;

    public ProjectConfig(final String projectName) throws IOException {
        this.projectName = projectName;

        // may fail:
        this.projectConf = FileUtils.getRequiredDirectory(Paths.DIR_CONFIG + projectName).getCanonicalFile();
        logger.info("Project Config: {}", projectConf);

        this.properties = loadProperties(projectConf);

        this.landingPageUrlMapping = loadUrlMapping(new File(this.projectConf, CONF_DOI_URL_MAPPING));

        this.templates = new DoiTemplates(
                new File(this.projectConf, Paths.DIR_PROJECT_TEMPLATES).getAbsolutePath()
        );

        initMetadataDir(
                new File(this.projectConf, Paths.DIR_PROJECT_METADATA).getAbsolutePath()
        );
        initInputDir(
                new File(this.projectConf, Paths.DIR_PROJECT_INPUTS).getAbsolutePath()
        );
        initTmpDir(
                new File(Paths.DIR_TMP, projectName).getAbsolutePath()
        );
        initTmpDoiDir(
                new File(tmpDir, Paths.DIR_TMP_DOI).getAbsolutePath()
        );

        initStagingDir(
                new File(Paths.DIR_STAGING, projectName).getAbsolutePath()
        );

        // /www/staging directories:
        initWebStagingDir(Paths.DIR_WEB_STAGING);
        initWebStagingProjectDir(
                new File(Paths.DIR_WEB_STAGING, projectName).getAbsolutePath()
        );
        initWebStagingProjectEmbedDir(
                new File(webStagingProjectDir, Paths.DIR_WEB_EMBED).getAbsolutePath()
        );
        initWebStagingProjectXmlDir(
                new File(webStagingProjectDir, Paths.DIR_WEB_XML).getAbsolutePath()
        );
        // /www/public directories:
        initWebPublicDir(Paths.DIR_WEB_PUBLIC);
        initWebPublicProjectDir(
                new File(Paths.DIR_WEB_PUBLIC, projectName).getAbsolutePath()
        );
    }

    public void initMetadataDir(final String dirPath) throws IOException {
        this.metadataCsvFiles = CsvUtil.findCSVFiles(dirPath);
    }

    // Used by junit test
    public void initInputDir(final String dirPath) throws IOException {
        this.inputCsvFiles = CsvUtil.findCSVFiles(dirPath);
    }

    public void initTmpDir(final String path) throws IOException {
        this.tmpDir = FileUtils.createDirectories(path);
    }

    public void initTmpDoiDir(final String path) throws IOException {
        this.tmpDoiDir = FileUtils.createDirectories(path);
    }

    public void initStagingDir(final String path) throws IOException {
        this.stagingDir = FileUtils.createDirectories(path);
    }

    public void initWebStagingDir(final String path) throws IOException {
        this.webStagingDir = FileUtils.createDirectories(path);
    }

    public void initWebStagingProjectDir(final String path) throws IOException {
        this.webStagingProjectDir = FileUtils.createDirectories(path);
    }

    public void initWebStagingProjectEmbedDir(final String path) throws IOException {
        this.webStagingProjectEmbedDir = FileUtils.createDirectories(path);
    }

    public void initWebStagingProjectXmlDir(final String path) throws IOException {
        this.webStagingProjectXmlDir = FileUtils.createDirectories(path);
    }

    public void initWebPublicDir(final String path) throws IOException {
        this.webPublicDir = FileUtils.createDirectories(path);
    }

    public void initWebPublicProjectDir(final String path) throws IOException {
        this.webPublicProjectDir = FileUtils.createDirectories(path);
    }

    public String getProjectName() {
        return projectName;
    }

    public File getProjectConf() {
        return projectConf;
    }

    public String getProperty(final String key) {
        final String val = properties.getProperty(key);
        if (val != null) {
            return val.trim();
        }
        return null;
    }

    public boolean getPropertyBoolean(final String key) {
        return Boolean.valueOf(properties.getProperty(key, "false"));
    }

    public String getLandingPageUrl(final String doiSuffix) {
        return landingPageUrlMapping.get(doiSuffix);
    }

    public DoiTemplates getTemplates() {
        return templates;
    }

    public File[] getMetadataCsvFiles() {
        return metadataCsvFiles;
    }

    public File[] getInputCsvFiles() {
        return inputCsvFiles;
    }

    public File getTmpDir() {
        return tmpDir;
    }

    public File getTmpDoiDir() {
        return tmpDoiDir;
    }

    public File getStagingDir() {
        return stagingDir;
    }

    public File getWebStagingDir() {
        return webStagingDir;
    }

    public File getWebStagingProjectDir() {
        return webStagingProjectDir;
    }

    public File getWebStagingProjectEmbedDir() {
        return webStagingProjectEmbedDir;
    }

    public File getWebStagingProjectXmlDir() {
        return webStagingProjectXmlDir;
    }

    public File getWebPublicDir() {
        return webPublicDir;
    }

    public File getWebPublicProjectDir() {
        return webPublicProjectDir;
    }

    public boolean isSaveCSV() {
        return saveCSV;
    }

    public void setSaveCSV(boolean saveCSV) {
        this.saveCSV = saveCSV;
    }

    private static Properties loadProperties(final File projectConf) {
        final File propFile = new File(projectConf, CONF_PROJECT);
        logger.info("Loading {}", propFile);

        final Properties props = new Properties();
        Reader r = null;
        try {
            r = FileUtils.reader(propFile);
            props.load(r);
        } catch (IOException ioe) {
            throw new RuntimeException("loadProperties failed:" + propFile.getAbsolutePath(), ioe);
        } finally {
            FileUtils.closeFile(r);
        }
        logger.info("properties: {}", props);
        return props;
    }

    private static Map<String, String> loadUrlMapping(final File csvFile) throws IOException {
        final Map<String, String> urlMapping;

        if (csvFile.exists()) {
            final CsvData data = CsvUtil.read(csvFile);
            logger.debug("CSV data: {}", data);

            final List<String[]> rows = data.getRows();

            urlMapping = new LinkedHashMap<String, String>(rows.size());

            // Get mappings [DOI|URL]
            for (ListIterator<String[]> it = rows.listIterator(); it.hasNext();) {
                final String[] cols = it.next();
                if (cols == null || cols.length != 2) {
                    // remove line:
                    it.remove();
                    continue;
                }
                String key = cols[0].trim();
                if (key.isEmpty() || key.charAt(0) == Const.COMMENT) {
                    continue;
                }
                String url = cols[1].trim();
                if (url.isEmpty() || !url.startsWith("http")) {
                    continue;
                }
                try {
                    // should detect repeated keys or values ?
                    urlMapping.put(key, new URL(url).toString());
                } catch (MalformedURLException mue) {
                    logger.info("invalid URL: {}", url);
                }
            }
        } else {
            urlMapping = Collections.emptyMap();
        }
        logger.info("urlMapping: {}", urlMapping);
        return urlMapping;
    }
}
