/*******************************************************************************
 * OSUG-DOI project ( http://doi.osug.fr ) - Copyright (C) CNRS.
 ******************************************************************************/
package fr.osug.doi.service;

import fr.osug.doi.domain.Doi;
import fr.osug.doi.domain.DoiPublic;
import fr.osug.doi.domain.DoiStaging;
import fr.osug.doi.domain.Project;
import fr.osug.doi.domain.Status;
import fr.osug.doi.repository.DoiPublicRepository;
import fr.osug.doi.repository.DoiRepository;
import fr.osug.doi.repository.DoiStagingRepository;
import fr.osug.doi.repository.ProjectRepository;
import java.util.Date;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * Service Implementation for managing Doi.
 */
@Service
public class DoiService {

    private final Logger log = LoggerFactory.getLogger(DoiService.class);

    @Autowired
    private ProjectRepository projectRepository;

    @Autowired
    private DoiRepository doiRepository;

    @Autowired
    private DoiStagingRepository doiStagingRepository;

    @Autowired
    private DoiPublicRepository doiPublicRepository;

    @Transactional
    public Project getOrCreateProject(final String name, final Date now) {
        Project project = projectRepository.findOneByName(name);

        if (project == null) {
            project = new Project();
            project.setName(name);

            project.setCreateDate(now);
            project.setUpdateDate(now);

            log.debug("saving: {}", project);

            projectRepository.save(project);
        }
        return project;
    }

    @Transactional
    public void updateProjectDate(final String name, final Date now) {
        Project project = projectRepository.findOneByName(name);

        if (project == null) {
            throw new IllegalArgumentException("Missing project with name: " + name);
        }

        project.setUpdateDate(now);

        log.debug("saving: {}", project);

        projectRepository.save(project);
    }

    @Transactional(readOnly = true)
    public Doi getDoi(final String identifier) {
        return doiRepository.findOneByIdentifier(identifier);
    }

    @Transactional
    public Doi createOrUpdateDoi(final Project project, final String identifier,
                                 final String description, final String landingExtUrl) {
        return createOrUpdateDoi(project, identifier, Status.STAGING, description, landingExtUrl);
    }

    @Transactional
    public Doi createOrUpdateDoi(final Project project, final String identifier,
                                 final Status status, final String description, final String landingExtUrl) {

        Doi doi = doiRepository.findOneByIdentifier(identifier);

        if (doi == null) {
            doi = new Doi();
            doi.setProject(project);
            doi.setIdentifier(identifier);
        }

        doi.setStatus(status);
        doi.setDescription(description);
        doi.setLandingExtUrl(landingExtUrl);

        log.debug("saving: {}", doi);

        doiRepository.save(doi);

        return doi;
    }

    @Transactional
    public DoiStaging createOrUpdateStagingDoi(final Project project, final String identifier,
                                               final String description, final String landingExtUrl,
                                               final boolean valid, final String metadataMd5,
                                               final String processLog,
                                               final Date now) {

        final Doi doi = createOrUpdateDoi(project, identifier, description, landingExtUrl);

        DoiStaging doiStaging = doiStagingRepository.findOneByDoi(doi);

        if (doiStaging == null) {
            doiStaging = new DoiStaging();
            doiStaging.setDoi(doi);

            doiStaging.setCreateDate(now);
            doiStaging.setUpdateDate(now);
        } else {
            // update date only if the MD5 changed:
            if (metadataMd5 != null && !metadataMd5.equals(doiStaging.getMetadataMd5())) {
                log.info("{} md5 modified: {}", doi.getIdentifier(), metadataMd5);
                doiStaging.setUpdateDate(now);
            }
        }

        doiStaging.setValid(valid);
        doiStaging.setMetadataMd5(metadataMd5);
        doiStaging.setLandingLocUrl("http://doi.osug.fr/staging/" + doi.getIdentifier()); // TODO
        doiStaging.setLog(processLog);

        log.debug("saving: {}", doiStaging);

        doiStagingRepository.save(doiStaging);

        return doiStaging;
    }

    @Transactional
    public DoiPublic createOrUpdatePublicDoi(final Project project, final DoiStaging doiStaging,
                                             final Date now) {
        final Doi doi = doiStaging.getDoi();

        DoiPublic doiPublic = doiPublicRepository.findOneByDoi(doi);

        if (doiPublic == null) {
            doiPublic = new DoiPublic();
            doiPublic.setDoi(doi);

            doiPublic.setCreateDate(now);
        }

        doiPublic.setMetadataMd5(doiStaging.getMetadataMd5());
        doiPublic.setLandingLocUrl("http://doi.osug.fr/public/" + doi.getIdentifier()); // TODO
        doiPublic.setActiveExtUrl(false); // TODO
        doiPublic.setUpdateDate(now);

        log.debug("saving: {}", doiPublic);

        doiPublicRepository.save(doiPublic);

        return doiPublic;
    }

    public ProjectRepository getProjectRepository() {
        return projectRepository;
    }

    public DoiRepository getDoiRepository() {
        return doiRepository;
    }

    public DoiStagingRepository getDoiStagingRepository() {
        return doiStagingRepository;
    }

    public DoiPublicRepository getDoiPublicRepository() {
        return doiPublicRepository;
    }

}
