/*******************************************************************************
 * OSUG-DOI project ( http://doi.osug.fr ) - Copyright (C) CNRS.
 ******************************************************************************/
package fr.osug.util;

import java.io.OutputStream;

/**
 * Compute the Seaman-Pence 32-bit 1's complement checksum over the byte stream EFFICIENTLY.
 * @author bourgesl
 */
public final class ChecksumOutputStream extends OutputStream {

    /* members */
    private boolean _open;
    /* partial sums */
    private long _hi;
    private long _lo;
    /* checksum */
    private long _checksum;

    ChecksumOutputStream() {
        reset();
    }

    void reset() {
        _open = true;
        _hi = 0l;
        _lo = 0l;
        _checksum = 0l;
    }

    long getChecksum() {
        /* close if needed to compute checksum */
        close();
        return _checksum;
    }

    @Override
    public void write(final int b) {
        throw new UnsupportedOperationException("Not supported: use write(byte b[], int off, int len).");
    }

    @Override
    public void write(byte data[], int offset, int length) {
        // check parameters:
        if (offset != 0) {
            throw new UnsupportedOperationException("Not supported: offset parameter must be 0.");
        }

        /*
             * Calculate the Seaman-Pence 32-bit 1's complement checksum over the byte stream. The option
             * to start from an intermediate checksum accumulated over another previous
             * byte stream is not implemented.
             * The implementation accumulates in two 64-bit integer values the two low-order and the two
             * high-order bytes of adjacent 4-byte groups. A carry-over of bits is never done within the main
             * loop (only once at the end at reduction to a 32-bit positive integer) since an overflow
             * of a 64-bit value (signed, with maximum at 2^63-1) by summation of 16-bit values could only
             * occur after adding approximately 140G short values (=2^47) (280GBytes) or more. We assume
             * for now that this routine here is never called to swallow FITS files of that size or larger.
             * @param data the byte sequence
             * @return the 32bit checksum in the range from 0 to 2^32-1
             * @see "http://heasarc.gsfc.nasa.gov/docs/heasarc/fits/checksum.html"
             * @author R J Mathar
             * @since 2005-10-05
         */
 /* get current state */
        long hi = _hi;
        long lo = _lo;

        final int len = length >> 1; // 2 * (length / 4)
        // System.out.println(length + " bytes") ;
        final int remain = length % 4;
        /* a write(2) on Sparc/PA-RISC would write the MSB first, on Linux the LSB; by some kind
             * of coincidence, we can stay with the byte order known from the original C version of
             * the algorithm.
         */
        for (int i = 0, j; i < len; i += 2) {
            /* The four bytes in this block handled by a single 'i' are each signed (-128 to 127)
                 * in Java and need to be masked indivdually to avoid sign extension /propagation.
             */
            j = i << 1;
            hi += (data[j] << 8) & 0xff00L | data[j + 1] & 0xffL;
            lo += (data[j + 2] << 8) & 0xff00L | data[j + 3] & 0xffL;
        }

        /* The following three cases actually cannot happen since FITS records are multiples of 2880 bytes.
         */
        if (remain >= 1) {
            hi += (data[2 * len] << 8) & 0xff00L;
        }
        if (remain >= 2) {
            hi += data[2 * len + 1] & 0xffL;
        }
        if (remain >= 3) {
            lo += (data[2 * len + 2] << 8) & 0xff00L;
        }

        /* update current state */
        _hi = hi;
        _lo = lo;
    }

    @Override
    public void close() {
        if (_open) {
            _open = false;
            /* get current state */
            long hi = _hi;
            long lo = _lo;

            long hicarry = hi >>> 16;
            long locarry = lo >>> 16;
            while (hicarry != 0 || locarry != 0) {
                hi = (hi & 0xffffL) + locarry;
                lo = (lo & 0xffffL) + hicarry;
                hicarry = hi >>> 16;
                locarry = lo >>> 16;
            }
            _checksum = (hi << 16) + lo;
        }
    }
}
